package ntaphttp

import (
	"fmt"
	"reflect"

	"github.com/caarlos0/env/v6"
	"github.com/sirupsen/logrus"
)

// DelayType is the type of delay
type DelayType string

const (
	// ConstantDelay is constant delay
	ConstantDelay DelayType = "Constant"
	// ExponentialDelay is exponential delay
	ExponentialDelay DelayType = "Exponential"
)

// IsDelayType returns if x is of DelayType enum
func IsDelayType(x DelayType) error {
	switch x {
	case ConstantDelay, ExponentialDelay:
		return nil
	default:
		return fmt.Errorf("%v is not of DelayType", x)
	}
}

// HTTPMethod is the http method
type HTTPMethod string

const (
	// GET is GET http method
	GET HTTPMethod = "GET"
	// PUT is PUT http method
	PUT HTTPMethod = "PUT"
	// POST is POST http method
	POST HTTPMethod = "POST"
	// PATCH is PATCH http method
	PATCH HTTPMethod = "PATCH"
)

// IsHTTPMethod returns if x is of HTTPMethod enum
func IsHTTPMethod(x HTTPMethod) error {
	switch x {
	case GET, PUT, POST, PATCH:
		return nil
	default:
		return fmt.Errorf("%v is not of HTTPMethod", x)
	}
}

// Config is the config for ntap http client
type Config struct {
	InsecureSkipVerify bool      `env:"INSECURE_SKIP_VERIFY,required"`
	RetryMaxTrial      int       `env:"RETRY_MAX_TRIAL,required"`
	RetryHTTPCodes     []int     `env:"RETRY_HTTP_CODES" envSeparator:","` // optional, use default if not found
	RequestTimeoutSec  int       `env:"REQUEST_TIMEOUT_SEC,required"`
	RetryDelayType     DelayType `env:"RETRY_DELAY_TYPE"` // optional, use constant delay if not found
	// constant delay time or base of exponential delay time
	RetryDelayBaseSec int          `env:"RETRY_DELAY_BASE_SEC,required"`
	RetryHTTPMethods  []HTTPMethod `env:"RETRY_HTTP_METHODS" envSeparator:","` // optional, use GET only if not found
	Logger            *logrus.Logger
}

// NewConfig returns an empty config
func NewConfig() *Config {
	return &Config{}
}

// NewConfigFromEnv returns config struct for http client
// return non pointer to make sure client does not modify
func NewConfigFromEnv() (*Config, error) {
	cfg := Config{}
	err := env.ParseWithFuncs(&cfg,
		// Convert string to specific types and validate enums
		map[reflect.Type]env.ParserFunc{
			reflect.TypeOf(GET): func(v string) (interface{}, error) {
				x := HTTPMethod(v)
				if err := IsHTTPMethod(x); err != nil {
					return nil, err
				}
				return x, nil
			},
			reflect.TypeOf(ConstantDelay): func(v string) (interface{}, error) {
				x := DelayType(v)
				if err := IsDelayType(x); err != nil {
					return nil, err
				}
				return x, nil
			},
		})

	if err != nil {
		return nil, fmt.Errorf("Fail to get retry HTTP client env vars: %v", err)
	}

	// if not specified, use default
	if len(cfg.RetryHTTPCodes) == 0 {
		cfg.RetryHTTPCodes = RetryableHTTPCodes
	}
	if len(cfg.RetryHTTPMethods) == 0 {
		cfg.RetryHTTPMethods = []HTTPMethod{GET}
	}
	if cfg.RetryDelayType == "" {
		cfg.RetryDelayType = ConstantDelay
	}

	return &cfg, nil
}

// WithInsecureSkipVerify sets InsecureSkipVerify in config
func (c *Config) WithInsecureSkipVerify(x bool) *Config {
	c.InsecureSkipVerify = x
	return c
}

// WithRetryMaxTrial sets RetryMaxTrial in config
func (c *Config) WithRetryMaxTrial(x int) *Config {
	c.RetryMaxTrial = x
	return c
}

// WithRetryHTTPCodes sets RetryHTTPCodes in config
func (c *Config) WithRetryHTTPCodes(x []int) *Config {
	c.RetryHTTPCodes = x
	return c
}

// WithRequestTimeMaxSec sets RequestTimeMaxSec in config
func (c *Config) WithRequestTimeMaxSec(x int) *Config {
	c.RequestTimeoutSec = x
	return c
}

// WithRetryDelayType sets DelayType in config
func (c *Config) WithRetryDelayType(x DelayType) *Config {
	c.RetryDelayType = x
	return c
}

// WithDelayBaseSec sets DelayBaseSec in config
func (c *Config) WithDelayBaseSec(x int) *Config {
	c.RetryDelayBaseSec = x
	return c
}

// WithRetryHTTPMethods sets RetryHTTPMethods in config
func (c *Config) WithRetryHTTPMethods(x []HTTPMethod) *Config {
	c.RetryHTTPMethods = x
	return c
}

func (c *Config) WithLogger(logger *logrus.Logger) *Config {
	c.Logger = logger
	return c
}
