package ntaphttp

import (
	"crypto/tls"
	"math"
	"net/http"
	"time"

	"github.com/PuerkitoBio/rehttp"
	"github.com/sirupsen/logrus"
	"github.com/thoas/go-funk"
)

// constDelay returns a DelayFn that always returns the same delay
// This function is copied from github.com/PuerkitoBio/rehttp to introduce logging
func constDelay(delay time.Duration, logger *logrus.Logger) rehttp.DelayFn {
	return func(attempt rehttp.Attempt) time.Duration {
		logger.Infof("Retrying with a delay: %v, Error: %v", delay, attempt.Error)
		return delay
	}
}

// expJitterDelay returns a DelayFn that returns a delay between 0 and
// base * 2^attempt capped at max (an exponential backoff delay with
// jitter).
// This function is copied from github.com/PuerkitoBio/rehttp to introduce logging
func expJitterDelay(base, max time.Duration, logger *logrus.Logger) rehttp.DelayFn {
	return func(attempt rehttp.Attempt) time.Duration {
		exp := math.Pow(2, float64(attempt.Index))
		top := float64(base) * exp
		delay := time.Duration(rehttp.PRNG.Int63n(int64(math.Min(float64(max), top))))
		logger.Infof("Retrying with a delay: %v, Error: %v", delay, attempt.Error)
		return delay
	}
}

// NewHTTPRetryClient constructs and returns a new
// http.Client with retry in transport layer which is used in default NewOntapClient function
// Using config.
// NOTE: this retry http client can be plugged in all swagger generated client
func NewHTTPRetryClient(cfg *Config) *http.Client {
	//Use default logger if config is not set with any logger
	if cfg.Logger == nil {
		cfg = cfg.WithLogger(logrus.New())
	}

	// create the transport to disable SSL verification
	// this transport implements RoundTripper interface
	httprt := &http.Transport{TLSClientConfig: &tls.Config{InsecureSkipVerify: cfg.InsecureSkipVerify}}

	// configure delay function
	delayFn := constDelay(time.Duration(cfg.RetryDelayBaseSec)*time.Second, cfg.Logger) // use constant delay by default
	if cfg.RetryDelayType == ExponentialDelay {
		delayFn = expJitterDelay(time.Duration(cfg.RetryDelayBaseSec)*time.Second,
			time.Duration(cfg.RequestTimeoutSec), cfg.Logger)
	}

	// Use rehttp as the underlying instead of http for retry
	tr := rehttp.NewTransport(
		httprt, // use constructed roundtripper
		rehttp.RetryAll(
			// Max retry attempts for a single request
			rehttp.RetryMaxRetries(cfg.RetryMaxTrial),
			rehttp.RetryHTTPMethods(
				funk.Map(cfg.RetryHTTPMethods, func(x HTTPMethod) string {
					return string(x)
				}).([]string)...),
			rehttp.RetryAny(
				// include retriable http code here
				rehttp.RetryStatuses(cfg.RetryHTTPCodes...),
				// retry all temporary errors, transport timeout, etc.
				// This is not well documented in Golang Transport package
				// TODO: test this case
				rehttp.RetryTemporaryErr(),
			),
		),
		delayFn,
	)

	httpclient := &http.Client{
		Transport: tr,
		Timeout:   time.Duration(cfg.RequestTimeoutSec) * time.Second, // for each request the max timeout
	}
	return httpclient
}
