package qstackcrypto

import (
	"crypto/aes"
	"crypto/cipher"
	"crypto/sha1"
	"encoding/base64"
	"fmt"
	"strings"

	"golang.org/x/crypto/pbkdf2"
)

const (
	separator = ":"
)

// SecretsHandler handles all cryptographic operations supported by qstackcrypto
type SecretsHandler struct {
	version *Version
	padder  *padder
}

// NewSecretsHandler returns an instance of SecretsHandler that can handle
// cryptographic operations using the specified version and block size
func NewSecretsHandler(version *Version) *SecretsHandler {
	return &SecretsHandler{
		version: version,
		padder:  newPadder(version),
	}
}

// Encrypt encrypts the specified plain text, using the specfied pass phrase,
// and returns the cypher text.
func (sh *SecretsHandler) Encrypt(passphrase, plaintext string) (*string, error) {
	salt, err := randomBytes(8)
	if err != nil {
		return nil, err
	}
	iv, err := randomBytes(16)
	if err != nil {
		return nil, err
	}

	key := pbkdf2.Key([]byte(passphrase), salt, sh.version.Iterations, sh.version.Bits/8, sha1.New)
	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}

	enctext := sh.padder.pad([]byte(plaintext))

	cbc := cipher.NewCBCEncrypter(block, iv)
	cbc.CryptBlocks(enctext, enctext)

	cyphertext := fmt.Sprintf("%s%s%s%s%s%s%s", sh.version.Name,
		separator, base64.StdEncoding.EncodeToString(salt),
		separator, base64.StdEncoding.EncodeToString(iv),
		separator, base64.StdEncoding.EncodeToString(enctext))
	return &cyphertext, nil
}

// Decrypt decrypts the specified cypher text, using the specified pass phrase,
// and returns the plain text.
func (sh *SecretsHandler) Decrypt(passphrase, cyphertext string) (*string, error) {
	if sh.version == nil {
		return nil, fmt.Errorf("Version not set for SecretsHandler")
	}
	if sh.padder.blockSize <= 0 {
		return nil, fmt.Errorf("Block size %v set for SecretsHandler is not valid", sh.padder.blockSize)
	}

	version, salt, iv, enctext, err := sh.parse(cyphertext)
	if err != nil {
		return nil, err
	}
	if version == nil || salt == nil || iv == nil || enctext == nil {
		return nil, fmt.Errorf("Internal error - one or more nil values returned from parsing")
	}
	if *version != sh.version.Name {
		return nil, fmt.Errorf("SecretsHandler not configured for specified cypher text")
	}

	key := pbkdf2.Key([]byte(passphrase), salt, sh.version.Iterations, sh.version.Bits/8, sha1.New)
	block, err := aes.NewCipher(key)
	if err != nil {
		return nil, err
	}

	cbc := cipher.NewCBCDecrypter(block, iv)
	cbc.CryptBlocks(enctext, enctext)

	plaintext, err := sh.padder.trim(enctext)
	if err != nil {
		return nil, err
	}
	return plaintext, nil
}

func (sh *SecretsHandler) parse(cyphertext string) (*string, []byte, []byte, []byte, error) {
	components := strings.Split(cyphertext, separator)
	if len(components) != 4 {
		return nil, nil, nil, nil, fmt.Errorf("Could not parse cypher text")
	}

	salt, err := base64.StdEncoding.DecodeString(components[1])
	if err != nil {
		return nil, nil, nil, nil, fmt.Errorf("Could not parse cypher text")
	}
	iv, err := base64.StdEncoding.DecodeString(components[2])
	if err != nil {
		return nil, nil, nil, nil, fmt.Errorf("Could not parse cypher text")
	}
	enctext, err := base64.StdEncoding.DecodeString(components[3])
	if err != nil {
		return nil, nil, nil, nil, fmt.Errorf("Could not parse cypher text")
	}

	return &components[0], salt, iv, enctext, nil
}
